from argparse import SUPPRESS
from getpass import getpass
import os
try:
    from ..utilities import schemas
except:
    import utilities.schemas as schemas


class Handlers:
    def fillout_form(self, form: list) -> dict:
        """
        fill out a form as requested by the server for more complicated functions
        """
        filled_form = dict()
        for field in form:
            value = str(input(f"{field}: ")).strip()
            if not value:
                value = None
            filled_form.update({field:value})
        return filled_form
    
    def form_handler(self, response):
        if not response.arguments_list:
            form = self.expression_input()
            filled_form = schemas.FormResponse(arguments_list=form)
        else: 
            form = self.fillout_form(response.arguments_list)
            filled_form = schemas.FormResponse(arguments_list=form)
        return filled_form

    def auth_handler(self, response):
        if response.secure_input or not response.requires_username:
            username = self.username
            password = getpass("Password: ")
        else: 
            username = input("Username: ")
            password = getpass("Password: ")
        filled_form = schemas.AuthData(username=username, password=password)
        return filled_form
    
    def confirmation_handler(self, response):
        print(response.message)
        while True:
            decision = str(input("(y/n)"))
            if decision == 'y':
                decision = True
                break
            elif decision == 'n':
                decision = False
                break
            else:
                print("Enter valid response")
        filled_form = schemas.ResponseData(response_message=decision)
        return filled_form
    
    def expression_input(self) -> str: 
        """
        Input an expression as requested by the server for something like cypher queries
        """
        print("Enter 'exit' to submit") 
        expression = ''
        line = ''
        while line != 'exit': 
            line = str(input("> "))
            if line != 'exit':
                expression += line
        return expression
    
    def environment_cli_response_stream_handler(self, response):
        if response.schema_type == 'ResponseData' and response.exit_status: # if the command was a shutdown or exit, close the program
            self.print_response(response.response_message)
            os._exit(0)
        elif response.schema_type == 'ResponseData':
            if response.active_username:
                self.username = response.active_username
            if response.url:
                self.url = response.url
            self.print_response(response.response_message)


from argparse import SUPPRESS
from getpass import getpass
import os
try:
    from ..utilities import schemas
except:
    import utilities.schemas as schemas


class Parsers:
    def process_command(self, command: str) -> list[str]: 
        """
        split the command string into a list. Not sure why this was even made
        """
        command = command.strip().split(' ') 
        return command

    def command_input_parser(self, kwargs: dict | list, exit_: int=0): 
        """
        parse arguments, handling bash and CLI input
        """
        if isinstance(kwargs, list): # check if the command input is from the CLI, or direct input
            kwargs = vars(self.parser.parse_args(kwargs)) 
        if not kwargs['command']:
            return False
        command = schemas.CommandData(kwargs = kwargs, exit_status = exit_)
        return command


class Args: # should add a way to automatically check for duplicate entries to this class
    """
    specifies the set of arguments for the argparser in the CLI.py file. Putting them in here allows the helpers.py file to access these options, and generate help menus
    """
    argparser_args = \
    {
        "id":{
            "args":["-i", "--id"],
            "kwargs":{"action":"store"}
        },
        "template":{
            "args":["-t", "--template"],
            "kwargs":{"action":"store"}
        },
        "username":{
            "args":["-u", "--username"],
            "kwargs":{"action":"store"}
        },
        "level":{
            "args":["-L", "--level"],
            "kwargs":{"action":"store"}
        },
        "version":{
            "args":["-V", "--version"],
            "kwargs":{"action":"store"}
        },
        "file":{
            "args":["-F", "--file"],
            "kwargs":{"action":"store"}
        },
        "name":{
            "args":["-n", "--name"],
            "kwargs":{"action":"store"}
        },
        "uuid":{
            "args":["-U", "--uuid"],
            "kwargs":{"action":"store"}
        },
        "link":{
            "args":["-l", "--link"],
            "kwargs":{"action":"store"}
        },
        "verbose":{
            "args":["-v", "--verbose"],
            "kwargs":{"action":"store_true"}
        },
        "help":{
            "args":['-h', '--help'],
            "kwargs":{"action":"store_true"}
        }
    }

    @staticmethod
    def check_duplicates():
        keys = []
        arguments = []
        for key, value in Args.argparser_args.items():
            if key in keys:
                raise AttributeError(f"Duplicate argument name {key} found in args")
            elif any(param in value['args'] for arg_list in arguments for param in arg_list): # horrific function I asked chatgpt to make because nested for loops are bad :(
                raise AttributeError(f"Duplicate command line parameter {value} foudn in args")
            keys.append(key)
            arguments.append(value)
            

if __name__ != "__main__":
    Args.check_duplicates



from tapipy import tapis
import pyperclip
import json
try:
    from . import baseCommand
    from ..utilities import decorators
except:
    import commands.baseCommand as baseCommand
    import utilities.decorators as decorators


class list_files(baseCommand.BaseCommand):
    """
    @help: list the files on a system 
    """
    async def run(self, id: str, file: str, *args, **kwargs) -> str: # lists files available on a tapis account
        file_list = self.t.files.listFiles(systemId=id, path=file)
        return str(file_list)
    

class upload(baseCommand.BaseCommand):
    """
    @help: upload a file to the system 
    the source and destination files must both be in the file argument, respectively, separated by a comma
    @todo: make it so that this doesnt need to take both source and destination files, but have it so it retrieves the current file location on the tapis system
    and sets that file location to be the upload point. Do the same for downloads but in reverse
    """
    async def run(self, file: str, id: str, *args, **kwargs) -> str: # upload a file from local to remote using tapis. Takes source and destination paths
        source = file.split(",")[0]
        destination = file.split(",")[1]
        self.t.upload(system_id=id,
                source_file_path=source,
                dest_file_path=destination)
        return f'successfully uploaded {source} to {destination}'


class download(baseCommand.BaseCommand):
    """
    @help: download a file from the system
    the source and destination files must both be in the file argument, respectively, separated by a comma
    """
    async def run(self, file: str, id: str, *args, **kwargs) -> str: # download a remote file using tapis, operates basically the same as upload
        source = file.split(",")[0]
        destination = file.split(",")[1]
        file_info = self.t.files.getContents(systemId=id,
                            path=source)

        file_info = file_info.decode('utf-8')
        with open(destination, 'w') as f:
            f.write(file_info)
        return f'successfully downloaded {source} to {destination}'


class Formatters:
    """
    Format received dictionaries in the client code
    """
    def recursive_dict_print(self, input_data: dict, depth: int=0):
        for key, value in input_data.items():
            if isinstance(value, dict):
                print(("  " * depth) + f"{key}:")
                self.recursive_dict_print(value, depth=depth + 1)
            elif isinstance(value, (list, tuple, set)):
                print(("  " * depth) + f"{key}:")
                for data in value:
                    print(("  " * (depth + 1)) + data)
            else: 
                print(("  " * depth) + f"{key}: {str(value).strip()}")
        if depth in (1, 0):
            print("\n")

    def print_response(self, response_message):
        """
        format response messages from the server
        """
        if type(response_message) == dict:
            self.recursive_dict_print(response_message)
        elif (type(response_message) == list or 
             type(response_message) == tuple or 
             type(response_message) == set):
            for value in response_message:
                self.print_response(value)
        else:
            print(response_message)


from tapipy import tapis
import pyperclip
try:
    from . import baseCommand
    from ..utilities import decorators
except:
    import commands.baseCommand as baseCommand
    import utilities.decorators as decorators


class get_pods(baseCommand.BaseCommand):
    """
    @help: return a list of pods the current tapis instance has access to
    @todo: add feature so that an individual pod id can be passed to access pod specific
    """
    async def run(self, *args, **kwargs) -> str: 
        pods_list = self.t.pods.get_pods()
        return pods_list

class create_pod(baseCommand.BaseCommand):
    """
    @help: create a new pod on the selected Tapis service
    """
    decorator = decorators.RequiresForm()
    async def run(self, id: str, template: str, 
                  description: str | None = None, *args, **kwargs) -> str:
        pod_information = self.t.pods.create_pod(pod_id=id, pod_template=template, description=description)
        return pod_information
    

class start_pod(baseCommand.BaseCommand):
    """
    @help: start the pod specified with ID
    """
    async def run(self, id: str, *args, **kwargs):
        return_information = self.t.pods.start_pod(pod_id=id)
        return return_information

class restart_pod(baseCommand.BaseCommand):
    """
    @help: initiate a pod restart
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, id: str, *args, **kwargs) -> str:
        return_information = self.t.pods.restart_pod(pod_id=id)
        return return_information


class stop_pod(baseCommand.BaseCommand):
    """
    @help: stop a pod's operations
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, id: str, *args, **kwargs):
        return_information = self.t.pods.stop_pod(pod_id=id)
        return return_information


class delete_pod(baseCommand.BaseCommand):
    """
    @help: delete select pod
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, id: str, *args, **kwargs) -> str: 
        return_information = self.t.pods.delete_pod(pod_id=id)
        return return_information
    

class set_pod_perms(baseCommand.BaseCommand):
    """
    @help: set the permissions for the pod selected
    """
    async def run(self, id: str, username: str, level: str, *args, **kwargs) -> str: # set pod permissions, given a pod id, user, and permission level
        return_information = self.t.pods.set_pod_permission(pod_id=id, user=username, level=level)
        return return_information


class delete_pod_perms(baseCommand.BaseCommand):
    """
    @help: delete the selected pod from the pods service you are connected to
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, id: str, username: str, *args, **kwargs) -> str: # take away someones perms if they are being malicious, or something
        return_information = self.t.pods.delete_pod_perms(pod_id=id, user=username)
        return return_information


class get_perms(baseCommand.BaseCommand):
    """
    @help: get the permissions list for the selected pod
    """
    async def run(self, id: str, *args, **kwargs) -> str: # return a list of permissions on a given pod
        return_information = self.t.pods.get_pod_permissions(pod_id=id)
        return return_information


class copy_pod_password(baseCommand.BaseCommand):
    """
    @help: copy the pod password to the clipboard
    """
    decorator=decorators.Auth()
    async def run(self, id: str, username=None, password=None, *args, **kwargs) -> str: # copies the pod password to clipboard so that the user can access the pod via the neo4j desktop app. Maybe a security risk? not as bad as printing passwords out!
        password = self.t.pods.get_pod_credentials(pod_id=id).user_password
        pyperclip.copy(password)
        password = None
        return 'copied to clipboard'


class get_pod_logs(baseCommand.BaseCommand):
    """
    @help: retrieve the logs of an active pod and either print them to the console, or write them to the specified file
    """
    async def run(self, id: str, file=None, *args, **kwargs):
        logs = self.t.pods.get_pod_logs(pod_id=id)
        if file:
            with open(file, 'w') as f:
                f.write(logs)
            return f"Log saved at {file}"
        return logs
    




import sys
from getpass import getpass
import time
import re
from tapipy.tapis import Tapis
import socket
import os
import logging
import typing
import asyncio
import traceback

try:
    from ..utilities import exceptions
    from ..utilities import logger
    from ..utilities import socketOpts as SO
    from ..utilities import killableThread
    from ..utilities import schemas
    from . import serverConnection
    from ..commands.query import neo4j, postgres
    from ..commands import baseCommand
    from ..commands import commandMap
    from ..utilities import decorators
except ImportError:
    import utilities.decorators as decorators
    import utilities.exceptions as exceptions
    import utilities.logger as logger
    import utilities.socketOpts as SO
    import utilities.killableThread as killableThread
    import utilities.schemas as schemas
    import server.serverConnection as serverConnection
    import commands.baseCommand as baseCommand
    import commands.commandMap as commandMap


class Server(commandMap.AggregateCommandMap, logger.ServerLogger, decorators.DecoratorSetup):
    """
    Receives commands from the client and executes Tapis operations
    """
    def __init__(self, IP: str, PORT: int):
        super().__init__()
        self.initial = True

        self.t = None
        self.url = None
        self.access_token = None
        self.username = None
        self.password = None

        self.__name__ = "Server"
        self.initialize_logger(self.__name__)
        # setting up socket server
        self.ip, self.port = IP, PORT
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setblocking(False)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.ip, self.port))
        self.sock.listen(1)
        self.end_time = time.time() + 300  # start the countdown on the timeout

        self.server = None

        self.logger.info('initialization complete')

    def switch_session(self, username: str, password, link: str, *args, **kwargs):
        start = time.time()
        try:
            t = Tapis(base_url=f"https://{link}",
                    username=username,
                    password=password)
            t.get_tokens()
        except Exception as e:
            self.logger.warning(e)
            raise ValueError(f"Invalid tapis auth credentials")
        
        self.username = username
        self.password = password
        self.t = t
        self.url = link
        self.access_token = self.t.access_token

        self.configure_decorators(self.username, self.password)
        self.update_credentials(t, username, password)
        # V3 Headers
        header_dat = {"X-Tapis-token": t.access_token.access_token,
                      "Content-Type": "application/json"}
        # create authenticator for tapis systems
        authenticator = t.access_token
        # extract the access token from the authenticator
        
        if 'win' in sys.platform:
            os.system(f"set JWT={self.access_token}")
        else: # unix based
            os.system(f"export JWT={self.access_token}")

        self.logger.info(f"initiated in {time.time()-start}")

        return f"Successfully initialized tapis service on {self.url}"

    async def handshake(self, connection):
        self.logger.info("Handshake starting")
        if self.initial:  # if this is the first time in the session that the cli is connecting
            startup_data = schemas.StartupData(initial = self.initial)
            await connection.send(startup_data)
            self.logger.info("send the initial status update")

            for attempt in range(1, 4):
                url: schemas.StartupData = await connection.receive()
                self.logger.info("received the link")
                try:
                    #await self.run_command(connection, {'command':'switch_service', 'link':url.url, 'verbose':False})
                    await self.aggregate_command_map['switch_service'](link=url.url, connection=connection, server=self, verbose=False)
                    self.logger.info("Verification success")
                    break
                except ValueError as e:
                    login_failure_data = schemas.ResponseData(response_message = (str(e), attempt))
                    await connection.send(login_failure_data)
                    self.logger.warning(f"Verification failure, {e}")
                    if attempt == 3:  
                        self.logger.error(
                            "Attempted verification too many times. Exiting")
                        raise ValueError("auth failure")
                    continue
        else:
            self.configure_decorators(self.username, self.password)
        self.initial = False
        startup_result = schemas.StartupData(initial = self.initial, username = self.username, url = self.url)
        self.logger.info("Connection success")
        await connection.send(startup_result)
        self.logger.info("Final connection data sent")

    async def accept(self, reader, writer):
        """
        accept connection request and initialize communication with the client
        """  
        connection = serverConnection.ServerConnection(reader=reader, writer=writer)
        self.timeout_handler()
        ip, port= writer.transport.get_extra_info('socket').getsockname()
        if ip != socket.gethostbyname(socket.gethostname()):
            raise exceptions.UnauthorizedAccessError(ip)
        self.logger.info("Received connection request")
        try:
            await self.handshake(connection)
        except ValueError:
            self.logger.warning("invalid credentials entered too many times. Cancelling request")
            await connection.close()
            return

        self.logger.info("connection is running now")
        loop = asyncio.get_event_loop()
        await loop.create_task(self.receive_and_execute(connection))


    def timeout_handler(self):  
        """
        checks if the timeout has been exceeded
        """
        if time.time() > self.end_time: 
            raise exceptions.TimeoutError

    async def receive_and_execute(self, connection):
        """
        receive and process commands
        """
        while True:
            try:
                message = await connection.receive()
                self.timeout_handler()  
                kwargs, exit_status = message.kwargs, message.exit_status
                result = await self.run_command(connection, kwargs)
                response = schemas.ResponseData(response_message = result, url = self.url, active_username = self.username)
                self.end_time = time.time() + 300 
                await connection.send(response)
                self.logger.info(message.schema_type)
            except (exceptions.TimeoutError, exceptions.Shutdown) as e:
                error_response = schemas.ResponseData(response_message = str(e), exit_status=1)
                await connection.send(error_response)
                self.logger.warning(str(e))
                self.server.close()
                loop = asyncio.get_event_loop()
                loop.stop()
                loop.close()
                sys.exit(0)
            except exceptions.Exit as e:
                self.logger.info("user exit initiated")
                error_response = schemas.ResponseData(response_message = str(e), exit_status=1)
                await connection.send(error_response)
                await connection.close()
                return
                #self.accept()  # wait for CLI to reconnect
            except OSError:
                self.logger.info("connection was lost, waiting to reconnect")
                await connection.close()
            except (exceptions.CommandNotFoundError, exceptions.NoConfirmationError, exceptions.InvalidCredentialsReceived, Exception) as e:
                error_str = traceback.format_exc()
                error_response = schemas.ResponseData(response_message = f"{str(e)}")
                await connection.send(error_response)
                self.logger.warning(f"{error_str}")
    
    async def main(self):
        self.server = await asyncio.start_server(self.accept, sock=self.sock)
        async with self.server:
            await self.server.serve_forever()


if __name__ == '__main__':
    server = Server(socket.gethostbyname(socket.gethostname()), 30000)
    asyncio.run(server.main())



import psycopg2
try:
    from ...utilities import decorators
    from .. import baseCommand
except:
    import utilities.decorators as decorators
    import commands.baseCommand as baseCommand


class postgres(baseCommand.BaseQuery):
    """
    @help: integrated CLI to interface with Postgres pods
    """
    decorator=decorators.RequiresExpression()
    async def run(self, id: str, expression: str=None, *args, **kwargs) -> str:
        uname, pword = self.get_pod_credentials(id)
        with psycopg2.connect(f"postgresql://{uname}:{pword}@{id}.pods.{self.t.base_url.split('https://')[1]}:443") as conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                cur.execute(query=expression)
                return_value = cur.fetchall()
        return str(f'[+][{id}] {return_value}')


#!/usr/bin/python3
import socket
import argparse
from argparse import SUPPRESS
import sys
import pyfiglet
from getpass import getpass
import os
import time
from client.handlers import Handlers
try:
    from ..utilities import schemas
    from ..utilities import socketOpts as SO
    from ..utilities import killableThread
    from ..utilities import decorators
    from ..utilities import args
    from ..utilities import logger
    from . import formatters
    from . import handlers
    from . import parsers
    from . import connectionInitializer
except:
    import utilities.schemas as schemas
    import utilities.socketOpts as SO
    import utilities.killableThread as killableThread
    import utilities.decorators as decorators
    import utilities.args as args
    import utilities.logger as logger
    import client.formatters as formatters
    import client.handlers as handlers
    import client.parsers as parsers
    import client.connectionInitializer as connectionInitializer


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
server_path = os.path.join(__location__, r'..\server\server.py')


class CLI(SO.ClientSocketOpts, decorators.DecoratorSetup, formatters.Formatters, args.Args, parsers.Parsers, connectionInitializer.ConnectionInitilializer):
    """
    Receive user input, either direct from bash environment or from the custom interface, then parse these commands and send them to the server to be executed. 
    """
    def __init__(self, IP: str, PORT: int):
        super().__init__()

        self.ip, self.port = IP, PORT
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 

        # sets up connection with the server
        self.username = None
        self.password = None

        # set up argparse
        self.parser = argparse.ArgumentParser(description="Command Line Argument Parser", exit_on_error=False, usage=SUPPRESS, conflict_handler='resolve')
        self.parser.add_argument('command')

        self.message_handlers = {
            'FormRequest':self.form_handler,
            'AuthRequest':self.auth_handler,
            'ConfirmationRequest':self.confirmation_handler,
        }

        for parameters in self.argparser_args.values():
            self.parser.add_argument(*parameters["args"], **parameters["kwargs"])

        print(r"If you find any issues, please create a new issue here: https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/issues")
    
    def special_forms_ops(self):
        """
        handle special form requests sent by the server
        """
        while True:
            message = self.schema_unpack_explicit(self.connection)
            message_type = message.schema_type
            if message_type in self.message_handlers.keys():
                filled_form = self.message_handlers[message_type](message)
            else:
                return message
            self.json_send_explicit(self.connection, filled_form.dict())

    def terminal_cli(self):
        self.username, self.url = self.connect()
        try:
            kwargs = self.parser.parse_args()
        except:
            print("Invalid Arguments")
            os._exit(0)
        kwargs = vars(kwargs)
        command = self.command_input_parser(kwargs, exit_=1) # operate with args, send them over
        self.json_send_explicit(self.connection, command.dict())
        response = self.special_forms_ops()
        if response.schema_type == 'ResponseData':
            self.print_response(response.response_message)
        os._exit(0)

    def environment_cli(self):
        self.username, self.url = self.connect()
        title = pyfiglet.figlet_format("-----------\nTapisCLICICLE\n-----------", font="slant") # print the title when CLI is accessed
        print(title)
        
        while True: # open the CLI if no arguments provided on startup
            try:
                time.sleep(0.01)
                kwargs = self.process_command(str(input(f"[{self.username}@{self.url}] "))) # ask for and process user input
                try:
                    command = self.command_input_parser(kwargs) # run operations
                except:
                    continue
                if not command:
                    continue
                self.json_send_explicit(self.connection, command.dict())
                response = self.special_forms_ops()
                self.environment_cli_response_stream_handler(response)
            except KeyboardInterrupt:
                pass # keyboard interrupts mess with the server, dont do it! it wont work anyway, hahahaha
            except WindowsError: # if connection error with the server (there wont be any connection errors)
                print("[-] Connection was dropped. Exiting")
                os._exit(0)
            except Exception as e: # if something else happens
                print(e)

    def main(self):
        if len(sys.argv) > 1: # checks if any command line arguments were provided. Does not open CLI
            self.terminal_cli()
        self.environment_cli()


if __name__ == "__main__":
    client = CLI(socket.gethostbyname(socket.gethostname()), 30000)
    client.main()


from typing import Callable, Optional, Union, Type
from abc import abstractmethod, ABC
import re
import inspect
import abc
from tapipy.tapis import TapisResult
from commands.dataFormatters import DataFormatters
try:
    from ..utilities import decorators
    from ..utilities import exceptions
    from ..utilities.args import Args
    from ..utilities import killableThread
except:
    import utilities.decorators as decorators
    import utilities.exceptions as exceptions
    import utilities.args as Args
    import utilities.killableThread as killableThread


EXCLUDED_ARGUMENTS = ('self', 'args', 'kwargs')


def get_kwargs(function):
    sig = inspect.signature(function)
    keyword_args = [param.name for param in sig.parameters.values() if param.default != inspect.Parameter.empty]
    return keyword_args

def get_args(function):
    sig = inspect.signature(function)
    positional_args = [param.name for param in sig.parameters.values() if param.default == inspect.Parameter.empty]
    return positional_args


class HelpStringRetriever:
    def help_string_retriever(self):
        try:
            docstring_components = self.__class__.__doc__
            docstring_components = docstring_components.split("@")
            for docstring_component in docstring_components:
                if re.match(r'^[^:]+', docstring_component).group(0) == "help":
                    return docstring_component.split("help: ")[1]
            raise Exception()
        except Exception as e:
            raise exceptions.HelpDoesNotExist(self.__class__.__name__)
    

class CommandMetaClass(abc.ABCMeta):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('BaseQuery', 'BaseCommand'):
            instance.__check_run(name, attrs)
            instance.__check_command_args(name, attrs)
            instance.__check_decorator(name, attrs)
        return instance
    
    def __check_run(self, name, attrs):
        if 'run' not in list(attrs.keys()):
            raise AttributeError(f"The command {name} requires a 'run()' method")
        elif not inspect.iscoroutinefunction(attrs['run']):
            raise AttributeError(f"The run method of the command {name} must be a coroutine")
    
    def __check_command_args(self, name, attrs):
        args = get_args(attrs['run'])
        for argument_name in args:
            if argument_name not in list(Args.Args.argparser_args.keys()) and argument_name not in EXCLUDED_ARGUMENTS:
                raise AttributeError(f"The argument {argument_name} in the run() method of the command '{name}' was not defined in the 'args' file")
        if 'kwargs' not in args or 'args' not in args:
            raise AttributeError(f"The run() method of the {name} class must have a **kwargs and *args attribute to ignore misinput")
        
    def __check_decorator(self,name, attrs):
        if 'decorator' in list(attrs.keys()) and type(attrs['decorator']) not in decorators.DECORATOR_LIST:
            raise AttributeError(f"The decorator parameter of the command {name} is invalid. Must be set to None or to a decorator type. Currently is {type(attrs['decorator'])}")


class BaseCommand(ABC, HelpStringRetriever, metaclass=CommandMetaClass):
    decorator = None
    return_formatter = None
    def __init__(self):
        self.t = None
        self.username = None
        self.password = None
        self.keyword_arguments = get_kwargs(self.run)
        self.positional_arguments = get_args(self.run)
        self.help = self.__help_gen()

    def set_t_and_creds(self, t, username, password):
        self.t = t
        self.username = username
        self.password = password
    
    def __argument_help_gen(self):
        arguments_str = ""
        for argument in self.positional_arguments:
            if argument not in EXCLUDED_ARGUMENTS:
                arguments_str += f" {Args.Args.argparser_args[argument]['args'][0]}/{Args.Args.argparser_args[argument]['args'][1]} <{argument}>"
        if not arguments_str:
            arguments_str = "No Arguments"
        return arguments_str
    
    def __help_gen(self):
        return {
            "Command":self.__class__.__name__,
            "Description":self.help_string_retriever(),
            "Syntax":f"{self.__class__.__name__} {self.__argument_help_gen()}"
        }

    @abstractmethod
    async def run(self, *args, **kwargs):
        pass

    async def __call__(self, **kwargs):
        if 'help' in list(kwargs.keys()) and kwargs['help']:
            return self.help
        if self.decorator:
            return_value = await self.decorator(self, **kwargs)
        else:
            return_value = await self.run(**kwargs)
        if self.return_formatter:
            return_value = self.return_formatter(return_value, kwargs['verbose'])
        return return_value
    

class BaseQuery(BaseCommand):
    decorator = None
    def get_pod_credentials(self, id):
        uname, pword = self.t.pods.get_pod_credentials(pod_id=id).user_username, self.t.pods.get_pod_credentials(pod_id=id).user_password
        return uname, pword
    
    @abstractmethod
    async def run(self, *args, **kwargs):
        pass
    

class CommandMapMetaClass(type):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('CommandMapMetaClass', 'BaseCommandMap'):
            instance.__check_commands_are_proper_type(name, attrs)
            instance.__check_command_name(name, attrs)
            instance.__check_data_formatter(name, attrs)
        return instance
    
    def __check_commands_are_proper_type(self, name, attrs):
        commands = attrs['command_map']
        if not commands:
            raise AttributeError(f"The command group {name} has no commands!")
        for command_name, command in commands.items():
            if not issubclass(command.__class__, BaseCommand):
                raise AttributeError(f"The command {command_name} was not passed to the {name} as a class, but as an object. Ensure you do not instantiate commands when defining the class")
            
    def __check_command_name(self, name, attrs):
        commands = attrs['command_map']
        for command_name, command in commands.items():
            if command_name != command.__class__.__name__:
                raise AttributeError(f"The command {command_name} in the command map {name} is a different name from its corresponding command class, {command.__class__.__name__}")
            
    def __check_data_formatter(self, name, attrs):
        if 'data_formatter' not in list(attrs.keys()):
            print(f"WARNING: The command map {name} has no data formatter. If any commands have non json-serializable return values, command execution will fail! These must be handled by a formatter")


class CommandContainer:
    aggregate_command_map: dict[str, Type[BaseCommand]] = dict()


class BaseCommandMap(CommandContainer, HelpStringRetriever, metaclass=CommandMapMetaClass):
    command_map: dict[str, Type[BaseCommand]] = None
    data_formatter = None
    def __init__(self):
        self.brief_help = self.__brief_help_gen()
        self.verbose_help = self.__help_gen()
        for name, command in self.command_map.items():
            command.return_formatter = self.data_formatter
            self.aggregate_command_map.update({name:command})

    def __help_gen(self):
        verbose_help = dict()
        for command in self.command_map.values():
            verbose_help.update({command.__class__.__name__:command.help})
        return verbose_help
    
    def __brief_help_gen(self):
        brief_help = {
            'group':self.__class__.__name__,
            'description':self.help_string_retriever(),
            'instruction':f"enter {self.__class__.__name__} to retrieve list of commands"
        }
        return brief_help
    
    def __call__(self):
        return self.verbose_help



from py2neo import Graph
try:
    from ...utilities import decorators
    from .. import baseCommand
except:
    import utilities.decorators as decorators
    import commands.baseCommand as baseCommand



class neo4j(baseCommand.BaseQuery):
    """
    @help: integrated CLI to interface with Neo4j pods
    """
    decorator=decorators.RequiresExpression()
    async def run(self, id: str, expression: str = None, *args, **kwargs) -> str: # function to submit queries to a Neo4j knowledge graph
        uname, pword = self.get_pod_credentials(id)
        graph = Graph(f"bolt+ssc://{id}.pods.{self.t.base_url.split('https://')[1]}:443", auth=(uname, pword), secure=True, verify=True)

        try:
            return_value = graph.run(expression)
            print(type(return_value))
            if str(return_value) == '(No data)' and 'create' in expression.lower(): # if no data is returned (mostly if something is created) then just say 'success'
                return f'[+][{id}@pods.{self.t.base_url.split("https://")[1]}:443] Success'
            elif str(return_value) == '(No data)':
                return f'[-][{id}@pods.{self.t.base_url.split("https://")[1]}:443] KG is empty'

            print(return_value)
            print(type(return_value))
            return str(f'[+][{id}] {return_value}')
        except Exception as e:
            return str(e)


"""
DECORATORS
These decorators are used in the tapisObjectWrappers.py file to standardize special functions. Allows for increased code reusability
"""
import typing
import enum
import sys
import time
import abc
import socket
from functools import update_wrapper, partial
try:
    from . import killableThread
    from . import schemas
    from . import socketOpts
    from . import exceptions
except:
    import killableThread
    import schemas
    import socketOpts
    import exceptions
    

class BaseRequirementDecorator(abc.ABC):
    username: typing.Optional[str] = None
    password: typing.Optional[str] = None

    @abc.abstractmethod
    async def __call__(self, command: typing.Type[object], connection: socket.socket, *args, **kwargs):
        pass


class RequiresForm(BaseRequirementDecorator):
    """
    This is for when you want to request separate input for specific command parameters instead of taking directly from the original command input (kwargs)
    Takes the parameters list of the function in question, filters out the ones that were not received from the original request message, sends another message 
    to request the unreceived parameters, and receives a message in response from the client to execute the function
    """
    async def __call__(self, command, *args, **kwargs):
        connection = kwargs['connection']
        if connection and not kwargs['file']:
            connection = connection
            fields = command.keyword_arguments
            if not fields:
                raise AttributeError(f"The decorated function {command} has no keyword parameters.")
            form_request = schemas.FormRequest(arguments_list=fields)
            await connection.send(form_request)
            filled_form: schemas.FormResponse = await connection.receive()
            for key, value in filled_form.arguments_list.items():
                kwargs[key] = value

        return await command.run(**kwargs)


class RequiresExpression(BaseRequirementDecorator):
    """
    This is for when you have something like a Neo4j or postgres interface to add to the tapisObjectWrappers file. Writing a Neo4j query directly in a command is cumbersome, its much
    easier to do if you have a blank, multiline environment to write. This will send a request for an expression, if an expression parameter exists in the decorated function.
    The client will open a new interface to type the expression. This is then sent back and fed to the function
    """
    async def __call__(self, command, *args, **kwargs):
        connection = kwargs['connection']
        if connection:
            connection = connection
            if 'expression' not in command.keyword_arguments:
                raise AttributeError(f"The function {command} does not contain an 'expression' keyword parameter")
            form_request = schemas.FormRequest(arguments_list=[])
            await connection.send(form_request)
            filled_form: schemas.FormResponse = await connection.receive()
            kwargs['expression'] = filled_form.arguments_list

        return await command.run(**kwargs)
    

class SecureInput(BaseRequirementDecorator):
    """
    Use this for functions where you need to hide input while typing into the cli. For instance, if you want to add a password to a service, as a user, but you dont actually
    want to authenticate. Checks if the decorated function has a password parameter, then requests secure input of a new password from the client
    """
    async def __call__(self, command, *args, **kwargs):
        connection = kwargs['connection']
        if connection:
            connection = connection
            if 'password' in command.keyword_arguments:
                secure_input_request = schemas.AuthRequest(secure_input=True)
                await connection.send(secure_input_request)
                secure_input_data: schemas.AuthData = await connection.receive()
                kwargs['password'] = secure_input_data.password
                return await command.run(**kwargs)
            raise AttributeError(f"The function {command} does not contain a 'password' parameter to securely input")
        return await command.run(**kwargs)


class Auth(BaseRequirementDecorator):
    """
    used for secure authentication from the client. Requires that the function has a username and password parameter for credentials. sends request for credentials from 
    the client, and checks those credentials against the stored credentials in the server.
    """
    async def __call__(self, command, *args, **kwargs):
        connection = kwargs['connection']
        requires_username = True
        if connection:
            if 'password' not in command.keyword_arguments:
                raise AttributeError(f"The command {command.__class__.__name__} does not have a 'password' keyword argument")
            if 'username' not in command.keyword_arguments:
                auth_request = schemas.AuthRequest(requires_username=False)
                requires_username = False
            else:
                auth_request = schemas.AuthRequest()
            await connection.send(auth_request)
            auth_data = await connection.receive()
            if auth_data.password != self.password and self.password:
                raise ValueError("The provided password does not match the stored password")
            if ((requires_username and auth_data.username != self.password) or (not requires_username and kwargs['username'] != self.username)) and self.username:
                raise ValueError("The provided username does not match the stored username")
            print(kwargs)
            kwargs['username'], kwargs['password'] = auth_data.username, auth_data.password
            return await command.run(**kwargs)
        return await command.run(**kwargs)


class NeedsConfirmation(BaseRequirementDecorator):
    """
    add to functions that you want user confirmation to exit. If you accidentally enter a command to delete a pod, this will not let you until you confirm
    """
    async def __call__(self, command, *args, **kwargs):
        connection = kwargs['connection']
        if connection:
            connection = connection
            confirmation_request = schemas.ConfirmationRequest(message=f"YOU REQUESTED TO {command.__class__.__name__}. THIS MIGHT CAUSE DATA LOSS! Please confirm (y/n)")
            await connection.send(confirmation_request)
            confirmation_reply: schemas.ResponseData = await connection.receive()
            confirmed = confirmation_reply.response_message
            if not confirmed:
                raise exceptions.NoConfirmationError(command)
        return await command.run(**kwargs)


DECORATOR_LIST = [
    NeedsConfirmation,
    RequiresForm,
    RequiresExpression,
    SecureInput,
    Auth
]
    
class DecoratorSetup:
    """
    for instantiation of the tapis wrappers, and the server, to set up decorators with user credentials and the socket connection. If you want to use the decorators in your class
    YOU WILL NEED TO USE THIS!
    """
    def configure_decorators(self, username, password):
        self.username = username
        self.password = password


class AnimatedLoading:
    """
    Add this if you want to print a loading animation while a function is executing
    """
    def __init__(self, func: typing.Callable):
        update_wrapper(self, func)
        self.func = func
        self.__code__ = func.__code__
        self.animation_frames = ['|','/','-','\\']

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)

    def __repr__(self):
        return self.func
    
    def __str__(self):
        return str(self.func)
    
    def animation(self):
        while True:
            for frame in self.animation_frames:
                print(f"loading {frame}", end='', flush=True)
                time.sleep(0.5)
    
    def __call__(self, obj, *args, **kwargs):
        animation_thread = killableThread.KillableThread(target=self.animation)
        animation_thread.start()
        result = self.func(obj, *args, **kwargs)
        animation_thread.kill()
        print("", end='', flush=True)
        return result
    

if __name__ == "__main__":
    print(issubclass(RequiresExpression, BaseRequirementDecorator))


import sys
import time
import re
from tapipy.tapis import Tapis
import os
import typing

try:
    from ..utilities import exceptions
    from ..utilities import decorators
    from ..utilities import args as Args
    from ..utilities import schemas
    from ..utilities import killableThread
    from . import baseCommand
except:
    import utilities.killableThread as killableThread
    import utilities.exceptions as exceptions
    import utilities.decorators as decorators
    import utilities.args as Args
    import utilities.decorators as decorators
    import utilities.schemas as schemas
    import commands.baseCommand as baseCommand


class switch_service(baseCommand.BaseCommand):
    """
    @help: switch the connected tapis service
    @todo: upgrade to federated auth
    """
    decorator = decorators.Auth()
    async def run(self, link: str, username: str=None, password=None, *args, **kwargs) -> tuple[typing.Any, str, str] | None:  # link is the baseURL
        results = kwargs['server'].switch_session(username, password, link)
        return results
      

class exit(baseCommand.BaseCommand):
    """
    @help: exit the CLI without shutting down the service
    """
    async def run(self, *args, **kwargs):
        raise exceptions.Exit
    

class shutdown(baseCommand.BaseCommand):
    """
    @help: exit the CLI and shutdown the service
    """
    decorator = decorators.NeedsConfirmation()
    async def run(self, *args, **kwargs):
        raise exceptions.Shutdown
    
class whoami(baseCommand.BaseCommand):
    """
    @help: returns the username of the current user
    """
    async def run(self, *args, **kwargs) -> str:
        user_info = self.t.authenticator.get_userinfo()
        return user_info
    

class get_args(baseCommand.BaseCommand):
    """
    @help: get the list of possible arguments 
    """
    async def run(self, *args, **kwargs):
        return Args.Args.argparser_args


from tapipy.tapis import TapisResult


HANDLED_TYPES = (TapisResult)
JSON_SERIALIZABLE = (str, int, dict, list, set, tuple)
    

class DataFormatters:
    pod_data = ('pod_id', 'pod_template', 'status')
    system_data = ('id', 'owner', 'systemType', 'canExec')
    app_data = ('id', 'owner', 'version', 'runtime')
    server_data = ('username',)

    @staticmethod
    def json_serializer(return_data):
        if isinstance(return_data, list):
            processed_data = []
            for data in return_data:
                processed_data.append(DataFormatters.json_serializer(data))
            return processed_data
        
        elif isinstance(return_data, dict):
            processed_data = dict()
            for name, data in return_data.items():
                processed_data[name] = DataFormatters.json_serializer(data)
            return processed_data

        elif isinstance(return_data, TapisResult):
            return DataFormatters.json_serializer(return_data.__dict__)
        
        return return_data

    @staticmethod
    def base_formatter(self, return_data, desired_data=None):
        filled_data = None
        if isinstance(return_data, list):
            filled_data = []
            for data in return_data:
                filled_data.append(DataFormatters.base_formatter(self, data, desired_data=desired_data))
            return filled_data
        elif isinstance(return_data, dict):
            filled_data = dict()
            for name, data in return_data.items():
                if name in desired_data:
                    filled_data[name] = data
            return filled_data
        return return_data
    
    @staticmethod
    def pod_formatter(self, return_data, verbose):
        if not verbose:
            return_data = DataFormatters.json_serializer(return_data)
            return DataFormatters.base_formatter(self, return_data, desired_data=DataFormatters.pod_data)
        return str(return_data)
    
    @staticmethod
    def system_formatter(self, return_data, verbose):
        if not verbose:
            return_data = DataFormatters.json_serializer(return_data)
            return DataFormatters.base_formatter(self, return_data, desired_data=DataFormatters.system_data)
        return str(return_data)
    
    @staticmethod
    def app_formatter(self, return_data, verbose):
        if not verbose:
            return_data = DataFormatters.json_serializer(return_data)
            return DataFormatters.base_formatter(self, return_data, desired_data=DataFormatters.app_data)
        return str(return_data)
    
    @staticmethod
    def server_formatter(self, return_data, verbose):
        if not verbose:
            return_data = DataFormatters.json_serializer(return_data)
            return DataFormatters.base_formatter(self, return_data, desired_data=DataFormatters.server_data)
        return str(return_data)
    



import client.handlers
import socket
import argparse
from argparse import SUPPRESS
import sys
import pyfiglet
from getpass import getpass
import os
import time
try:
    from ..utilities import schemas
    from ..utilities import socketOpts as SO
    from ..utilities import killableThread
    from ..utilities import decorators
    from ..utilities import args
    from ..utilities import logger
    from . import formatters
    from . import handlers
    from . import parsers
except:
    import utilities.schemas as schemas
    import utilities.socketOpts as SO
    import utilities.killableThread as killableThread
    import utilities.decorators as decorators
    import utilities.args as args
    import utilities.logger as logger
    import client.formatters as formatters
    import client.handlers as handlers
    import client.parsers as parsers


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
server_path = os.path.join(__location__, r'..\serverRun.py')
print(server_path)


class ConnectionInitilializer(handlers.Handlers):
    def initialize_server(self): 
        """
        detect client operating system. The local server intitialization is different between unix and windows based systems
        """
        if 'win' in sys.platform:
            os.system(f"pythonw {server_path}")
        else: # unix based
            os.system(f"python {server_path} &")

    #@decorators.AnimatedLoading
    def connection_initialization(self):
        """
        start the local server through the client
        """
        startup_flag = False
        timeout_time = time.time() + 30 # server setup timeout. If expires, there is a problem!
        while True:
            if time.time() > timeout_time: # connection timeout condition
                sys.stdout.write("\r[-] Connection timeout")
                os._exit(0)
            try:
                self.connection.connect((self.ip, self.port)) 
                if startup_flag:
                    startup.kill()
                break
            except Exception as e:
                if not startup_flag:
                    startup = killableThread.KillableThread(target=self.initialize_server) # run the server setup on a separate thread
                    startup.start() 
                    startup_flag = True # set the flag to true so the thread runs only once
                    continue

    def connect(self):
        """
        connect to the local server
        """
        self.connection_initialization() 
        #self.connection.connect((self.ip, self.port)) # enable me for debugging. Requires manual server start
        connection_info: schemas.StartupData = self.schema_unpack_explicit(self.connection) # receive info from the server whether it is a first time connection
        if connection_info.initial: # if the server is receiving its first connection for the session\
            while True:
                try:
                    url = str(input("\nEnter the uri for the tapis service you are connecting to: ")).strip()
                except KeyboardInterrupt:
                    url = " "
                    pass
                url_data = schemas.StartupData(url=url)
                self.json_send_explicit(self.connection, url_data.dict())
                auth_request: schemas.AuthRequest = self.schema_unpack_explicit(self.connection)
                while True:
                    try:
                        auth_data = self.auth_handler(auth_request)
                        break
                    except KeyboardInterrupt:
                        pass
                self.json_send_explicit(self.connection, auth_data.dict())

                verification: schemas.ResponseData | schemas.StartupData = self.schema_unpack_explicit(self.connection)
                if verification.schema_type == 'StartupData': # verification success, program moves forward
                    return verification.username, verification.url
                else: # verification failed. User has 3 tries, afterwards the program will shut down
                    print(f"[-] verification failure, attempt # {verification.response_message[1]}")
                    if verification.response_message[1] == 3:
                        sys.exit(0)
                    continue

        return connection_info.username, connection_info.url # return the username and url


from typing import Callable, Optional, Union, Type
from abc import abstractmethod, ABC
import re
import inspect
from tapipy import tapis
try:
    from ..utilities import exceptions
    from . import baseCommand, serverCommands, fileCommands, podCommands, appCommands, systemCommands, dataFormatters
    from .query import neo4j, postgres
except ImportError:
    import utilities.exceptions as exceptions
    import commands.baseCommand as baseCommand
    import commands.serverCommands as serverCommands
    import commands.fileCommands as fileCommands
    import commands.podCommands as podCommands
    import commands.appCommands as appCommands
    import commands.systemCommands as systemCommands
    import commands.query.neo4j as neo4j
    import commands.query.postgres as postgres
    import commands.dataFormatters as dataFormatters


class Systems(baseCommand.BaseCommandMap):
    """
    @help: run operations on Tapis systems
    """
    data_formatter = dataFormatters.DataFormatters.system_formatter
    command_map = {
        'get_systems':systemCommands.get_systems(), # since initialization of commands is separate from __init__, you dont need to specify these as classes anymore
        'get_system_info':systemCommands.get_system_info(),
        'create_system':systemCommands.create_system(),
        'set_system_credentials':systemCommands.set_system_credentials(),
        'set_system_password':systemCommands.set_system_password(),
        'delete_system':systemCommands.delete_system(),
    }


class Server(baseCommand.BaseCommandMap):
    """
    @help: run config operations on the 
    """
    data_formatter = dataFormatters.DataFormatters.server_formatter
    command_map = {
        'whoami':serverCommands.whoami(),
        'exit':serverCommands.exit(),
        'shutdown':serverCommands.shutdown(),
        "get_args":serverCommands.get_args(),
        'switch_service':serverCommands.switch_service()
    }


class Pods(baseCommand.BaseCommandMap):
    """
    @help: run operations on tapis pods
    """
    data_formatter = dataFormatters.DataFormatters.pod_formatter
    command_map = {
        'get_pods':podCommands.get_pods(),
        'create_pod':podCommands.create_pod(),
        'start_pod':podCommands.start_pod(),
        'restart_pod':podCommands.restart_pod(),
        'delete_pod':podCommands.delete_pod(),
        'set_pod_perms':podCommands.set_pod_perms(),
        'stop_pod':podCommands.stop_pod(),
        'delete_pod_perms':podCommands.delete_pod_perms(),
        'get_perms':podCommands.get_perms(),
        'copy_pod_password':podCommands.copy_pod_password(),
        'get_pod_logs':podCommands.get_pod_logs(),
    }


class Files(baseCommand.BaseCommandMap):
    """
    @help: run operations on tapis files
    """
    command_map = {
        'list_files':fileCommands.list_files(),
        'upload':fileCommands.upload(),
        'download':fileCommands.download(),
    }


class Apps(baseCommand.BaseCommandMap):
    """
    @help: Run operations on tapis apps
    """
    data_formatter = dataFormatters.DataFormatters.app_formatter
    command_map = {
        'create_app':appCommands.create_app(),
        'get_apps':appCommands.get_apps(),
        'delete_app':appCommands.delete_app(),
        'get_app':appCommands.get_app(),
        'run_job':appCommands.run_job(),
        'get_job_status':appCommands.get_job_status(),
        'download_job_output':appCommands.download_job_output(),
    }


class Query(baseCommand.BaseCommandMap):
    """
    @help: run integrated query CLIs
    """
    command_map = {
        'postgres': postgres.postgres(),
        'neo4j': neo4j.neo4j()
    }


class AggregateCommandMap(baseCommand.CommandContainer):
    groups = {
        'Systems': Systems(),
        'Server': Server(),
        'Pods': Pods(),
        'Files': Files(),
        'Apps': Apps(),
        'Query': Query(),
    }
    def __init__(self):
        self.help = self.__general_help()

    def __general_help(self):
        general_help = list()
        for group in self.groups.values():
            general_help.append(group.brief_help)
        return general_help

    def update_credentials(self, t, username, password):
        for command_name, command in self.aggregate_command_map.items():
            command.set_t_and_creds(t, username, password)

    async def run_command(self, connection, command_data: dict):
        """
        process and run command based on received kwargs
        """
        command_data['connection'] = connection
        command_data['server'] = self
        print(f"FROM RUN COMMAND {command_data}")
        command_name = command_data['command']
        if command_name in list(self.aggregate_command_map.keys()):
            command = self.aggregate_command_map[command_name]
            command_data.pop('command')
            return await command(**command_data)
        elif command_name in list(self.groups.keys()):
            return self.groups[command_name]()
        elif command_name == "help":
            return self.help
        else:
            raise exceptions.CommandNotFoundError(command_name)


from tapipy import tapis
import pyperclip
import json
import os
try:
    from . import baseCommand
    from ..utilities import decorators
except:
    import commands.baseCommand as baseCommand
    import utilities.decorators as decorators


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
server_path = os.path.join(__location__, '..\\server.py')


class get_systems(baseCommand.BaseCommand):
    """
    @help: Gets and returns the list of systems the current Tapis service and account have access to
    @doc: this is an example of the doc segment of the docstring. not included in help message
    """
    async def run(self, *args, **kwargs):
        systems = self.t.systems.getSystems()
        return systems
    

class get_system_info(baseCommand.BaseCommand):
    """
    @help: get information on a selected system
    """
    async def run(self, id:str, *args, **kwargs): 
        system_info = self.t.systems.getSystem(systemId=id)
        return system_info
    

class set_system_credentials(baseCommand.BaseCommand):
    """
    @help: upload system credentials to a system. Must generate keys first using 'ssh-keygen -m PEM -f id_rsa', and format with, 'awk -v ORS='\\n' '1' <private_key_name>
    file argument must contain the path to the private and public keys respectively, separated by a ','
    """
    async def run(self, id: str, file: str, *args, **kwargs): # get information about a system given its ID
        with open(file.split(",")[0], 'r') as f:
            private_key = f.read()

        with open(file.split(",")[1], 'r') as f:
            public_key = f.read()

        cred_return_value = self.t.systems.createUserCredential(systemId=id,
                            userName=self.username,
                            privateKey=private_key,
                            publicKey=public_key)

        return str(cred_return_value)


class create_system(baseCommand.BaseCommand):
    """
    @help: create a system. Must have a properly configured system file.
    see the template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/system-config.json
    this command will automatically create and upload the ssh keys
    """
    set_system_creds = set_system_credentials()
    def __keygen(self):
        local_files = os.listdir(__location__)
        if "id_rsa" not in local_files or "id_rsa.pub" not in local_files:
            os.system(f'ssh-keygen -q -m PEM -f {__location__}\\id_rsa -N ""')
            with open(f"{__location__}\\id_rsa", 'r') as f:
                formatted_key = ""
                for line in f.readlines()[1:-1]:
                    formatted_key += line.strip()

            with open(f"{__location__}\\id_rsa", 'w') as f:
                f.write(formatted_key)

    async def run(self, file: str, *args, **kwargs) -> str: # create a tapius system. Takes a path to a json file with all system information, as well as an ID
        self.set_system_creds.set_t_and_creds(self.t, self.username, self.password)

        self.__keygen()
        with open(file, 'r') as f:
            system = json.loads(f.read())
        return_value = self.t.systems.createSystem(**system)
        self.set_system_creds(id=system['id'], file=f"{__location__}\\id_rsa,{__location__}\\id_rsa.pub")
        return return_value
    

class set_system_password(baseCommand.BaseCommand):
    """
    @help: set a system password
    """
    decorator=decorators.SecureInput()
    async def run(self, id: str, password: str=None, *args, **kwargs) -> str: # set the password for a system
        try:
            password_return_value = self.t.systems.createUserCredential(systemId=id, # will put this in a getpass later
                                userName=self.username,
                                password=password)
            return str(password_return_value)
        except Exception as e:
            raise Exception(f"{e}\nTry running set_credentials if the problem persists")
        

class delete_system(baseCommand.BaseCommand):
    """
    @help: delete the selected system
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, id: str, *args, **kwargs) -> str:
        return_value = self.t.systems.deleteSystem(systemId=id)
        return return_value
